searchData={"items":[{"type":"module","title":"Flint","doc":"Declarative [`Ecto`](https://github.com/elixir-ecto/ecto) `embedded_schema`s for data validation, coercion, and manipulation.","ref":"Flint.html"},{"type":"module","title":"Feature Overview - Flint","doc":"* `!` Variants of Ecto `field`, `embeds_one`, and `embeds_many` macros to mark a field as required ([Required Fields](#required-fields))\n* Colocated input transformations let you either transform input fields before validation or derive field values from other fields ([Derived Fields / Input Transformations](#derived-fields--input-transformations))\n* Colocated validations, so you can define common validations alongside field declarations ([Validations](#field-validations))\n* Colocated output transformations let you transform fields after validation ([Mappings / Output Transformations](#mappings--output-transformations))\n* Extensible using the `Flint.Extension` module. Default extensions include:\n  * `Accessible` - Adds `Access` implementation to the target schemas\n  * `JSON` - Adds a custom JSON encoding (`Jason` and `Poison` supported) implementation to the target schemas\n  * `Embedded` - Sets good default module attribute values used by `Ecto` specifically tailored for in-memory embedded schemas\n  * And more!\n* New [`Ecto.Schema` Reflection Functions](https://hexdocs.pm/ecto/Ecto.Schema.html#module-reflection)\n  * `__schema__(:required)` - Returns list of fields marked as required (from `!` macros)\n  * And more!\n* Convenient generated function (`changeset`,`new`,`new!`,...) ([Generated Functions](#generated-functions))\n* Configurable `Application`-wide defaults for `Ecto.Schema` API ([Config](#config))\n* Conveniently create new `Ecto` types using the `Flint.Type` module and its  `deftype/2` macro ([`Flint.Type`](#flinttype))","ref":"Flint.html#module-feature-overview"},{"type":"module","title":"Installation - Flint","doc":"```elixir\ndef deps do\n  [\n    {:flint, \"~> 0.4\"}\n  ]\nend\n```","ref":"Flint.html#module-installation"},{"type":"module","title":"Motivation - Flint","doc":"`Flint` is built on top of Ecto and is meant to provide good defaults for using `embedded_schema`s for use outside of a database.\nIt also adds a bevy of convenient features to the existing `Ecto` API to make writing schemas and validations much quicker.\n\nOf course, since you're using Ecto, you can use this for use as an ORM, but this is emphasizing the use of `embedded_schema`s as just more expressive and powerful maps while keeping compatibility with `Ecto.Changeset`, `Ecto.Type`, and all of the other benefits Ecto has to offer.\n\nIn particular, `Flint` focuses on making it more ergonomic to use `embedded_schema`s as a superset of Maps, so a `Flint.Schema` by default implements the `Access` behaviour and implements the `Jason.Encoder` protocol.\n\n`Flint` also was made to leverage the distinction `Ecto` makes between the embedded representation of the schema and the dumped representation. This means that you can dictate how you want the Elixir-side representation to look, and then provide transformations\nfor how it should be dumped, which helps when you want the serialized representation to look different.\n\nThis is useful if you want to make changes in the server-side code without needing to change the client-side (or vice-versa). Or perhaps you want a mapped representation, where instead of an `Ecto.Enum` just converting its atom key to a string when dumped, it gets mapped to an integer, etc.","ref":"Flint.html#module-motivation"},{"type":"module","title":"Basic Usage - Flint","doc":"If you want to declare a schema with `Flint`, just `use Flint.Schema` within your module, and now you have access to `Flint`'s implementation of the\n`embedded_schema/1` macro.  You can declare an `embedded_schema` within your module as you otherwise would with `Ecto`. Within the `embedded_schema/1` block, you also have access to `Flint`s implementations of `embeds_one`,`embeds_one!`,`embeds_many`, `embeds_many!`, `field`, and `field!`.\n\n```elixir\ndefmodule User do\n  use Flint.Schema\n\n  embedded_schema do\n    field! :username, :string\n    field! :password, :string, redacted: true\n    field :nickname, :string\n  end\nend\n```\n\n## `Flint` Types\n\n`Flint` also comes with some types that are automatically aliased when you `use Flint`.\n\n### `Union`\n\nUnion type for Ecto. Allows the field to be any of the specified types.\n\n## `Flint.Type`\n\n`Flint.Type` is meant to make writing new `Ecto` types require much less boilerplate, because you can base your type off of an existing type and only modify the callbacks that have different behavior.\n\nSimply `use Flint.Type` and pass the `:extends` option which says which type module to inherit callbacks\nfrom.  This will delegate all required callbacks and any implemented optional callbacks and make them\noverridable.\n\nIt also lets you make a type from an `Ecto.ParameterizedType` with default parameter values.\nYou may supply any number of default parameters. This essentially provides a new\n`init/1` implementation for the type, supplying the default values, while not affecting any of the\nother `Ecto.ParameterizedType` callbacks. You may still override the newly set defaults at the local level.\n\nJust supply all options that you wish to be defaults as extra options when using `Flint.Type`.\n\nYou may override any of the inherited callbacks inherity from the extended module\nin the case that you wish to customize the module further.\n\n#","ref":"Flint.html#module-basic-usage"},{"type":"module","title":"Examples - Flint","doc":"``` elixir\ndefmodule Category do\n  use Flint.Type, extends: Ecto.Enum, values: [:folder, :file]\nend\n```\n\nThis will apply default `values` to `Ecto.Enum` when you supply a `Category` type\nto an Ecto schema. You may still override the values if you supply the `:values`\noption for the field.\n\n```elixir\nimport Flint.Type\ndeftype NewUID, extends: Ecto.UUID, dump: &String.length/1\n```\n\nThis will create a new `NewUID` type that behaves exactly like an `Ecto.UUID` except it dumps\nits string length.","ref":"Flint.html#module-examples"},{"type":"module","title":"Generated Functions - Flint","doc":"`Flint` provides default implementations for the following functions for any schema declaration. Each of these is overridable.\n\n* `changeset` - Creates a changeset by casting all fields and validating all that were marked as required. If a `:default` key is provided for a field, then any use of a bang (`!`) declaration will essentially be ignored since the cast will fall back to the default before any validations are performed.\n* `new` - Creates a new changeset from the empty module struct and applies the changes (regardless of whether the changeset was valid).\n* `new!` - Same as new, except raises if the changeset is not valid.","ref":"Flint.html#module-generated-functions"},{"type":"module","title":"Flint Core - Flint","doc":"The core of Flint is the additional schema macros, which includes the bang (`!`) variants to mark\na field as required, and the added support of validations through `do` blocks to fields, as well\nas the `Flint.Extension` API that allows extensions to define additional acceptable `field` options\nand module attributes that can be reflected upon.\n\nAll other functionality comes in the form of `Flint` extensions.\n\nAt their core, the new `field` and `field!` macros' only additional functionality over the default `Ecto` macros is to store the allowed `Flint` options into module attributes which are exposed as new reflection functions.\n\nThe bulk of the work done in Flint with validations and transformations of data occurs in the generated `changeset` function, which leaves it up to the end user whether to use the default implementation, roll their own from scratch using the information exposed through the reflection functions, or do something in between (such as tuning which extensions you use).\n\nWhen you `use Flint.Schema`, you declare an overridable `changeset` function for your schema module that by default just\ndelegates to the `Flint.Pipeline.changeset/3` function.\n\nThe `Flint.Pipeline.changeset/3` function operates as the following pipeline:\n\n1. Cast all fields (including embeds)\n2. Validate required fields ([Required Fields](#required-fields))\n\n#","ref":"Flint.html#module-flint-core"},{"type":"module","title":"Required Fields - Flint","doc":"`Flint` adds the convenience bang (`!`) macros (`embed_one!`,`embed_many!`, `field!`) for field declarations within your struct to declare a field as required within its `changeset` function.\n\n`Flint` schemas also have a new reflection function in addition to the normal [`Ecto` reflection functions](https://hexdocs.pm/ecto/Ecto.Schema.html#module-reflection).\n\n* `__schema__(:required)` -- Returns a list of all fields that were marked as required.\n\n#","ref":"Flint.html#module-required-fields"},{"type":"module","title":"Field Validations - Flint","doc":"#### `field` `do` Blocks\n\nIn `Flint`, the `field` and `field!` macros now accept an optional `do` block to define condition/error pairs.\n\n```elixir\n  embedded_schema do\n    field! :type, :string do\n      type not in ~w[elf human] -> \"Expected elf or human, got: #{type}\"\n    end\n\n    field! :age, :integer do\n      age < 0 ->\n        \"Nobody can have a negative age\"\n\n      type == \"elf\" and age > max_elf_age ->\n        \"Attention! The elf has become a bug! Should be dead already!\"\n\n      type == \"human\" and age > max_human_age ->\n        \"Expected human to have up to #{max_human_age}, got: #{age}\"\n    end\n  end\n```\n\n```elixir\nmax_elf_age = 400\nmax_human_age = 120\nCharacter.new!(%{type: \"elf\", age: 10}, binding())\n```\n\n```elixir\n** (ArgumentError) %Character{type: [\"Expected elf or human, got: orc\"], age: 10}\n    (flint 0.0.1) lib/schema.ex:617: Flint.Schema.new!/3\n    (elixir 1.15.7) src/elixir.erl:396: :elixir.eval_external_handler/3\n    (stdlib 5.1.1) erl_eval.erl:750: :erl_eval.do_apply/7\n    (elixir 1.15.7) src/elixir.erl:375: :elixir.eval_forms/4\n    (elixir 1.15.7) lib/module/parallel_checker.ex:112: Module.ParallelChecker.verify/1\n    lib/livebook/runtime/evaluator.ex:622: anonymous fn/3 in Livebook.Runtime.Evaluator.eval/4\n    (elixir 1.15.7) lib/code.ex:574: Code.with_diagnostics/2\n```\n\nThe `:do` block accepts a list of validation clauses, where is clause is of the form:\n\n`failure condition -> Error Message`\n\nIn the `:do` block expressions, the same rules apply as mentioned across this documentation. You can pass\nbindings to apply to the expression, and field name bindings will automatically be passed to the expression so you can just use the field names as variables.\n\nAdditionally, you will have access to all local functions and imported functions that the parent module would have, so you can write expressions as you would in the parent module.\n\nThe AST representations of these vaildations are stored in a module attribute and can be retrieved using\n\n```elixir\n__schema__(:blocks)\n```\n\nThese validations are checked in the default `Flint.Pipeline.changeset`.\n\n##","ref":"Flint.html#module-field-validations"},{"type":"module","title":"Parameterized Validations - Flint","doc":"You can even parameterize the options passed to the validations:\n\n```elixir\ndefmodule Person do\n  use Flint\n\n  embedded_schema do\n    field! :first_name, :string,  max: 10, min: 5\n    field! :last_name, :string, min: 5, max: 10\n    field :favorite_colors, {:array, :string}, subset_of: [\"red\", \"blue\", \"green\"]\n    field! :age, :integer, greater_than: 0, less_than: max_age\n  end\nend\n```\n\nIf you do this, make sure to pass the options as a keyword list into the call to `changeset`:\n\n```elixir\nPerson.changeset(\n  %Person{},\n  %{first_name: \"Bob\", last_name: \"Smith\", favorite_colors: [\"red\", \"blue\", \"pink\"], age: 101},\n  [max_age: 100]\n)\n```\n\n```elixir\n#Ecto.Changeset ,\n  valid?: false,\n  ...\n>\n```\n\n##","ref":"Flint.html#module-parameterized-validations"},{"type":"module","title":"Validate With Respect to Other Fields - Flint","doc":"You might find yourself wishing to validate a field conditionally based on the values of other fields. In `Flint`, you\ncan do this with any validation! Since all validations already accept parameterized conditions, they also let you refer\nto previously defined fields declared with `field` or `field!` macros. Just use a variable of the same name as the field(s) you want to refer to, and they will be bound to their respective variables.","ref":"Flint.html#module-validate-with-respect-to-other-fields"},{"type":"module","title":"Extensions - Flint","doc":"Flint provides an extensible architecture using the `Flint.Extension` module.\n\nIn fact, most of the core features that `Flint` offers are written as `Flint` extensions.\n\nThe default extensions can be retrieved using `Flint.default_extensions()`\n\n```elixir\nFlint.default_extensions()\n\n[\n  Flint.Extensions.PreTransforms,\n  Flint.Extensions.When,\n  Flint.Extensions.EctoValidations,\n  Flint.Extensions.PostTransforms,\n  Flint.Extensions.Accessible,\n  Flint.Extensions.Embedded,\n  Flint.Extensions.JSON\n]\n```\n\nTo use extensions, you can specify them when using `Flint.Schema` in your module. If the `:extensions` option\nis not provided, the default extensions will be used.\n\n```elixir\ndefmodule MySchema do\n  use Flint.Schema,\n    extensions: [Accessible, JSON]\n\n  embedded_schema do\n    # Schema fields...\n  end\nend\n```\n\n**Note that you don't have to write the fully-qualified name for modules in the `Flint.Extensions` module.**\n\nYou can use `Flint.default_extensions()` to refer to the default extensions, which you will have to\nexplicitly add if you pass custom values to the `:extensions` option when using `Flint.Schema`, eg.\n\n```elixir\ndefmodule MySchema do\n  use Flint.Schema,\n    extensions: Flint.default_extensions() ++ [MyExtension]\n\n  embedded_schema do\n    # Schema fields...\n  end\nend\n```\n\nYou can also create custom extensions by `use`ing `Flint.Extension`. This allows you to add additional functionality or modify the behavior of Flint schemas according to your specific needs.\n\nFor more details on creating and using extensions, refer to the `Flint.Extension` module documentation.","ref":"Flint.html#module-extensions"},{"type":"module","title":"Aliases - Flint","doc":"If you don't like the name of an option, you can provide a compile-time list of aliases to map new option names to existing options\n([Validation Options](#basic-validation-options) and [Transformation Options](#input-and-output-transformations)).\n\nIn your config, add an `:aliases` key with a `Keyword` value, where each key is the new alias, and the value is an existing option name.\n\nFor example, these are default aliases implemented in `Flint`:\n\n```elixir\nconfig Flint, aliases: [\n    lt: :less_than,\n    gt: :greater_than,\n    le: :less_than_or_equal_to,\n    ge: :greater_than_or_equal_to,\n    eq: :equal_to,\n    ne: :not_equal_to\n  ]\n```\n\n**NOTE** If you add your own aliases and want to keep these above defaults, you will have to add them manually to your aliases.","ref":"Flint.html#module-aliases"},{"type":"module","title":"Config - Flint","doc":"You can configure the default options set by `Flint`.\n\n* `embeds_one`: The default arguments when using `embeds_one`. Defaults to `[defaults_to_struct: true, on_replace: :delete]`\n* `embeds_one!`: The default arguments when using `embeds_one!`. Defaults to `[on_replace: :delete]`\n* `embeds_many`: The default arguments when using `embeds_many` or `embeds_many!`. Defaults to `[on_replace: :delete]`\n* `embeds_many!`: The default arguments when using `embeds_many!`. Defaults to `[on_replace: :delete]`\n* `:enum`: The default arguments for an `Ecto.Enum` field. Defaults to `[embed_as: :dumped]`.\n* `:aliases`: [Aliases](#aliases)\n\nYou can also configure any aliases you want to use for schema validations.","ref":"Flint.html#module-config"},{"type":"module","title":"Embedded vs Dumped Representations - Flint","doc":"`Flint` takes advantage of the distinction `Ecto` makes between an `embedded_schema`'s embedded and dumped representations.\n\nFor example, by default in `Flint`, `Ecto.Enum`s that are `Keyword` (rather than just lists of atoms) will have their keys\nbe the embedded representation and will have the values be the dumped representation.\n\n```elixir\ndefmodule Book do\n  use Flint.Schema, schema: [\n    field(:genre, Ecto.Enum, values: [biography: 0, science_fiction: 1, fantasy: 2, mystery: 3])\n  ]\nend\n\nbook = Book.new(%{genre: \"biography\"})\n# %Book{genre: :biography}\n\nFlint.Schema.dump(book)\n# %{genre: 0}\n```\n\nIn this example, you can see how you can share multiple representations of the same data using this distinction.\n\nYou can also implement your own `Ecto.Type` and further customize this (see [`Flint.Type`](#flinttype)):\n\n```elixir\ndefmodule ContentType do\n  use Ecto.Type\n  def type, do: :atom\n\n  def cast(\"application/json\"), do: {:ok, :json}\n\n  def cast(_), do: :error\n  def load(_), do: :error\n\n  def dump(:json), do: {:ok, \"application/json\"}\n  def dump(_), do: :error\n\n  def embed_as(_) do\n    :dump\n  end\nend\n```\n\nHere, `cast` will be called when creating a new `Flint` schema from a map, and `dump` will be used\nwhen calling `Flint.Schema.dump/1`.\n\n```elixir\ndefmodule URL do\n  use Flint, schema: [\n    field(:content_type, ContentType)\n  ]\nend\n```\n\n```elixir\nurl = URL.new(%{content_type: \"application/json\"})\n# %URL{content_type: :json}\n\nFlint.Schema.dump(url)\n# %{content_type: \"application/json\"}\n```","ref":"Flint.html#module-embedded-vs-dumped-representations"},{"type":"module","title":"Examples - Flint","doc":"You can view the [Notebooks folder](https://github.com/acalejos/flint/tree/main/notebooks) for some examples in Livebook.\n\nYou can also look at [Merquery](https://github.com/acalejos/merquery/tree/main/lib/merquery/schemas) for a real, comprehensive\nexample of how to use `Flint`.","ref":"Flint.html#module-examples"},{"type":"function","title":"Flint.default_extensions/1","doc":"","ref":"Flint.html#default_extensions/1"},{"type":"module","title":"Flint.Changeset","doc":"The base `changeset` function defined by `Flint`. `Flint.Changeset` uses the module attributes\nthat are collected when using the `Flint.Schema` macros to perform transformations and validations.","ref":"Flint.Changeset.html"},{"type":"function","title":"Flint.Changeset.changeset/3","doc":"Given a `Flint` (or `Ecto`) schema and params (can be a map, struct of the given schema, or an existing changeset),\napplies all steps of the `Flint.Changeset` to generate a new changeset.\n\nThis function casts all fields (recursively casting all embeds using this same function),\nvalidates required fields (specified using the bang (`!`) macros exposed by `Flint`),\noutputting the resulting `Ecto.Changeset`.","ref":"Flint.Changeset.html#changeset/3"},{"type":"function","title":"Flint.Changeset.validate_do_blocks/2","doc":"Uses the quoted expressions from the `Flint.Schema.field` and `Flint.Schema.field!`\n`do` blocks to validate the changeset.\n\nYou can optionally pass bindings to be added to the evaluation context.","ref":"Flint.Changeset.html#validate_do_blocks/2"},{"type":"module","title":"Flint.Schema","doc":"`Flint.Schema` provides custom implementations of certain `Ecto` `embedded_schema` DSL keywords.\n\nWhen you `use Flint`, all of these definitions are imported into the module and override the default\n`Ecto.Schema` implementations. You should not have to directly interact with this module.","ref":"Flint.Schema.html"},{"type":"function","title":"Flint.Schema.dump/1","doc":"Dumps the JSON representation of the given schema","ref":"Flint.Schema.html#dump/1"},{"type":"macro","title":"Flint.Schema.embedded_schema/1","doc":"Wraps `Ecto`'s `embedded_schema` macro, injecting `Flint`'s custom macro implementation into the module space.","ref":"Flint.Schema.html#embedded_schema/1"},{"type":"macro","title":"Flint.Schema.embeds_many/3","doc":"Wraps `Ecto`'s `embeds_many` macro to accept additional options which are consumed by `Flint.Changeset`.\n\n`Flint` options that are passed to `embeds_many` are stripped by `Flint` before passing them to `Ecto.Schema.embeds_many` and\nstored in module attributed for the schema's module.\n\nThe following default options are passed to `Flint.Schema.embeds_many` and can be overriden at the `Application` level or at the\nlocal call level.","ref":"Flint.Schema.html#embeds_many/3"},{"type":"macro","title":"Default Options - Flint.Schema.embeds_many/3","doc":"```elixir\non_replace: :delete\n```","ref":"Flint.Schema.html#embeds_many/3-default-options"},{"type":"macro","title":"Flint.Schema.embeds_many/4","doc":"","ref":"Flint.Schema.html#embeds_many/4"},{"type":"macro","title":"Flint.Schema.embeds_many!/3","doc":"Marks an embeds_many field as required by storing metadata in a module attribute, then calls `Flint.Schema.embeds_many`.\n\nThe metadata tracking is not enforced at the schema / struct level (eg. through enforced keys), but rather\nat the schema validation level (through something such as `Flint.Changeset.changeset`).\n\nThe following default options are passed to `Flint.Schema.embeds_many!` and can be overriden at the `Application` level or at the\nlocal call level.","ref":"Flint.Schema.html#embeds_many!/3"},{"type":"macro","title":"Default Options - Flint.Schema.embeds_many!/3","doc":"```elixir\non_replace: :delete\n```","ref":"Flint.Schema.html#embeds_many!/3-default-options"},{"type":"macro","title":"Flint.Schema.embeds_many!/4","doc":"","ref":"Flint.Schema.html#embeds_many!/4"},{"type":"macro","title":"Flint.Schema.embeds_one/3","doc":"Wraps `Ecto`'s `embeds_one` macro to accept additional options which are consumed by `Flint.Changeset`.\n\n`Flint` options that are passed to `embeds_one` are stripped by `Flint` before passing them to `Ecto.Schema.embeds_one` and\nstored in module attributed for the schema's module.\n\nThe following default options are passed to `Flint.Schema.embeds_one` and can be overriden at the `Application` level or at the\nlocal call level.","ref":"Flint.Schema.html#embeds_one/3"},{"type":"macro","title":"Default Options - Flint.Schema.embeds_one/3","doc":"```elixir\ndefaults_to_struct: true,\non_replace: :delete\n```","ref":"Flint.Schema.html#embeds_one/3-default-options"},{"type":"macro","title":"Flint.Schema.embeds_one/4","doc":"","ref":"Flint.Schema.html#embeds_one/4"},{"type":"macro","title":"Flint.Schema.embeds_one!/3","doc":"Marks an embeds_one field as required by storing metadata in a module attribute, then calls `Flint.Schema.embeds_one`.\n\nThe metadata tracking is not enforced at the schema / struct level (eg. through enforced keys), but rather\nat the schema validation level (through something such as `Flint.Changeset.changeset`).\n\nThe following default options are passed to `Flint.Schema.embeds_one!` and can be overriden at the `Application` level or at the\nlocal call level.","ref":"Flint.Schema.html#embeds_one!/3"},{"type":"macro","title":"Default Options - Flint.Schema.embeds_one!/3","doc":"```elixir\non_replace: :delete\n```","ref":"Flint.Schema.html#embeds_one!/3-default-options"},{"type":"macro","title":"Flint.Schema.embeds_one!/4","doc":"","ref":"Flint.Schema.html#embeds_one!/4"},{"type":"macro","title":"Flint.Schema.field/3","doc":"Wraps `Ecto`'s `field` macro to accept additional options which are consumed by `Flint.Changeset`.\n\n`Flint` options that are passed to `field` are stripped by `Flint` before passing them to `Ecto.Schema.field` and\nstored in module attributed for the schema's module.\n\nIf no `Flint`-specific options or features are used, this acts the same as `Ecto.Schema.field`.","ref":"Flint.Schema.html#field/3"},{"type":"macro","title":"Flint.Schema.field/4","doc":"","ref":"Flint.Schema.html#field/4"},{"type":"macro","title":"Flint.Schema.field!/3","doc":"Marks a field as required by storing metadata in a module attribute, then calls `Flint.Schema.field`.\n\nThe metadata tracking is not enforced at the schema / struct level (eg. through enforced keys), but rather\nat the schema validation level (through something such as `Flint.Changeset.changeset`).","ref":"Flint.Schema.html#field!/3"},{"type":"macro","title":"Flint.Schema.field!/4","doc":"","ref":"Flint.Schema.html#field!/4"},{"type":"function","title":"Flint.Schema.new/3","doc":"Creates a new schema struct according to the schema's `changeset` implementation, immediately applying\nthe changes from the changeset regardless of errors.\n\nIf you want to manually handle error cases, you should use the `changeset` function itself.","ref":"Flint.Schema.html#new/3"},{"type":"function","title":"Flint.Schema.new!/3","doc":"Same as `new`, except will `raise` if any errors exist in the changeset.","ref":"Flint.Schema.html#new!/3"},{"type":"module","title":"Flint.Type","doc":"`Flint.Type` is meant to make writing new `Ecto` types require much less boilerplate, because you can base your\ntype off of an existing type and only modify the callbacks that have different behavior.\n\nSimply `use Flint.Type` and pass the `:extends` option which says which type module to inherit callbacks\nfrom.  This will delegate all required callbacks and any implemented optional callbacks and make them\noverridable.\n\nIt also lets you make a type from an `Ecto.ParameterizedType` with default parameter values.\nYou may supply any number of default parameters. This essentially provides a new\n`init/1` implementation for the type, supplying the default values, while not affecting any of the\nother `Ecto.ParameterizedType` callbacks. You may still override the newly set defaults at the local level.\n\nJust supply all options that you wish to be defaults as extra options when using `Flint.Type`.\n\nYou may override any of the inherited callbacks inherity from the extended module\nin the case that you wish to customize the module further.","ref":"Flint.Type.html"},{"type":"module","title":"Examples - Flint.Type","doc":"``` elixir\ndefmodule Category do\n  use Flint.Type, extends: Ecto.Enum, values: [:folder, :file]\nend\n```\n\nThis will apply default `values` to `Ecto.Enum` when you supply a `Category` type\nto an Ecto schema. You may still override the values if you supply the `:values`\noption for the field.\n\n```elixir\nimport Flint.Type\ndeftype NewUID, extends: Ecto.UUID, dump: &String.length/1\n```\n\nThis will create a new `NewUID` type that behaves exactly like an `Ecto.UUID` except it dumps\nits string length.","ref":"Flint.Type.html#module-examples"},{"type":"macro","title":"Flint.Type.deftype/2","doc":"A shorthand for creating a new `Flint.Type`. This is equivalent to\ncreating a new module with name `module` and calling `use Flint.Type` and passing `opts`.","ref":"Flint.Type.html#deftype/2"},{"type":"macro","title":"Example - Flint.Type.deftype/2","doc":"```elixir\ndeftype Vehicle, extends: Ecto.Enum, values: [:car, :motorcycle, :truck]\n```","ref":"Flint.Type.html#deftype/2-example"},{"type":"module","title":"Flint.Types.Union","doc":"Union type for Ecto. Allows the field to be any of the specified types.\n\nThis is particularly useful if you want to allow one of multiple types and preserve the type of the\ninput, so long as it adheres to the type (rather than forcing a cast).\n\nFor example, suppose you want to allow an input to be a float, or an integer, but you want to\npreserve whichever type is passed (keep floats as floats and keep integers as integers), you could\nuse a `Union`.","ref":"Flint.Types.Union.html"},{"type":"module","title":"Options - Flint.Types.Union","doc":"* `:oneof` - Allowed types. Can be any valid `Ecto` type (including custom types). Required.\n* `:eager` - Whether to eagerly cast the field value into the `:oneof` in the order that\nthey appear in the list. If `false`, `Flint` will attempt to determine if the value maps to a\nvalid `Ecto` type before casting and will prefer that type (if it appears in `:oneof`). Defaults\nto `false`.","ref":"Flint.Types.Union.html#module-options"},{"type":"module","title":"Example - Flint.Types.Union","doc":"```elixir\ndefmodule Character do\n  use Flint\n\n  embedded_schema do\n    field! :name, :string\n    field! :strength, Union, oneof: [:integer, :float]\n  end\nend\n```\n\n```elixir\nCharacter.new!(%{name: \"Elf\", strength: 10})\n```\n\n```\n%Character{name: \"Elf\", strength: 10}\n```\n\n```elixir\nCharacter.new!(%{name: \"Elf\", strength: 10.0})\n```\n\n```\n%Character{name: \"Elf\", strength: 10.0}\n```","ref":"Flint.Types.Union.html#module-example"},{"type":"module","title":"Flint.Extension","doc":"`Flint` extensions allow developers to easily hook into `Flint` metaprogramming lifecycle to add extra data into the embedded\nschema reflection functions.\n\nFlint currently offers three ways to extend behavior:\n\n1. Schema-level attributes\n2. Field-level additional options\n3. Injected Code\n\nExtension authors define what fields / options / attributes `Flint` should look for in the module / schema definition and\nstrip out and store in a schema reflection function, but it is still the resposibility of either the extension author or\nthe end user to make use of the stored information.","ref":"Flint.Extension.html"},{"type":"module","title":"Schema-Level Attributes - Flint.Extension","doc":"These are simply module attributes that are pre-registered with `Flint`, and can be given a default value\nas well as a validation function. When you use an extension that registers an attribute, then a new `__schema__`\nreflection function is added for each attribute name, with the attribute name as the argument.\n\n**Note that the validation occurs at compile time**\n\n#","ref":"Flint.Extension.html#module-schema-level-attributes"},{"type":"module","title":"Example - Flint.Extension","doc":"Given the following extension:\n\n```elixir\ndefmodule Returnable do\n  use Flint.Extension\n\n  attribute :returns, validator: fn returns -> is_binary(returns) end\nend\n```\n\nAnd the schema\n\n```elixir\ndefmodule Schema do\n  use Flint.Schema, extensions: [Returnable]\n  @returns \"something\"\n  embedded_schema do\n    ...\n  end\nend\n```\n\nThen you can reflect on this new attribute with:\n\n```elixir\nSchema.__schema__(:returns)\n```","ref":"Flint.Extension.html#module-example"},{"type":"module","title":"Field-Level Options - Flint.Extension","doc":"You can also register additional field-level keyword options to be consumed in a downstream function.\n\nThese function similarly to the built-in extra options that `Flint` provides, where the options are\nstripped and stored in a module attribute (and subsequently in a `__schema__` reflection function)\nbefore passing the valid `Ecto.Schema` options to `Ecto` itself.\n\n**Note that the validation occurs at compile time**\n\n#","ref":"Flint.Extension.html#module-field-level-options"},{"type":"module","title":"Example - Flint.Extension","doc":"Given the following extension that enables Go-like JSON marshalling options:\n\n```elixir\ndefmodule JSON do\n  use Flint.Extension\n\n  option :name, required: false, validator: &is_binary/1\n  option :omitempty, required: false, default: false, validator: &is_boolean/1\n  option :ignore, required: false, default: false, validator: &is_boolean/1\nend\n```\n\nAnd the following schema:\n\n```elixir\ndefmodule Schema do\n  use Flint.Schema, extensions: [JSON]\n  embedded_schema do\n    field :myfield, :string, name: \"my_field\", omitempty: true\n  end\nend\n```\n\nThen you can access these specific fields with:\n\n```elixir\nSchema.__schema__(:extra_options)\n```\n\n```elixir\n[\n  myfield: [ignore: false, omitempty: true, name: \"my_field\"],\n]\n```","ref":"Flint.Extension.html#module-example"},{"type":"module","title":"Injected Code - Flint.Extension","doc":"Lastly, extensions allow you to define custom `__using__/1` macros which will be passed through\nto the target schema module. This is one of the core functionalities of extensions, and works the\nsame as you would normally `use` a module, and helps compartmentalize similar functionality.","ref":"Flint.Extension.html#module-injected-code"},{"type":"module","title":"Default Extensions - Flint.Extension","doc":"By default, `Flint` will enable the following extensions:\n\n*  `Flint.Extensions.PreTransforms`,\n*  `Flint.Extensions.When`,\n*  `Flint.Extensions.EctoValidations`,\n*  `Flint.Extensions.PostTransforms`,\n*  `Flint.Extensions.Accessible`,\n*  `Flint.Extensions.Embedded`,\n*  `Flint.Extensions.JSON`\n\nIf you want to pass your own list of extensions for a module, you will need to explicitly pass the defaults\nas well if you would like to keep them. You can use the convenience `Flint.default_extensions/0` constant\nif you want to include all of the defaults.\n\n\n#","ref":"Flint.Extension.html#module-default-extensions"},{"type":"module","title":"Options - Flint.Extension","doc":"* `:extensions` (list of module that adopts `Spark.Dsl.Extension`) - A list of DSL extensions to add to the `Spark.Dsl`\n\n* `:otp_app` (`t:atom/0`) - The otp_app to use for any application configurable options\n\n* `:fragments` (list of `t:module/0`) - Fragments to include in the `Spark.Dsl`. See the fragments guide for more.","ref":"Flint.Extension.html#module-options"},{"type":"module","title":"Flint.Extensions.Accessible","doc":"An extension to automatically implement the `Access` behaviour for your struct,\ndeferring to the `Map` implementation.","ref":"Flint.Extensions.Accessible.html"},{"type":"module","title":"Flint.Extensions.EctoValidations","doc":"Shorthand options for common validations found in `Ecto.Changeset`\n\nJust passthrough the option for the appropriate validation and this extension\nwill take care of calling the corresponding function from `Ecto.Changeset` on\nyour data.","ref":"Flint.Extensions.EctoValidations.html"},{"type":"module","title":"Options - Flint.Extensions.EctoValidations","doc":"* `:greater_than` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n* `:less_than` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n* `:less_than_or_equal_to` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n* `:greater_than_or_equal_to` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n* `:equal_to` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n* `:not_equal_to` ([`Ecto.Changeset.validate_number/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_number/3-options))\n* `:format` ([`Ecto.Changeset.validate_format/4`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_format/4))\n* `:subset_of` ([`Ecto.Changeset.validate_subset/4`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_subset/4))\n* `:in` ([`Ecto.Changeset.validate_inlusion/4`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_inclusion/4))\n* `:not_in` ([`Ecto.Changeset.validate_exclusion/4`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_exclusion/4))\n* `:is` ([`Ecto.Changeset.validate_length/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_length/3-options))\n* `:min` ([`Ecto.Changeset.validate_length/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_length/3-options))\n* `:max` ([`Ecto.Changeset.validate_length/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_length/3-options))\n* `:count` ([`Ecto.Changeset.validate_length/3`](https://hexdocs.pm/ecto/Ecto.Changeset.html#validate_length/3-options))\n* `:when` - Let's you define an arbitrary boolean condition on the field which can refer to any `field` defined above it or itself. **NOTE** The `:when` option will output a generic error on failure, so if verbosity is desired, an [advanced validation](#advanced-validations) is more appropriate.","ref":"Flint.Extensions.EctoValidations.html#module-options"},{"type":"module","title":"Aliases - Flint.Extensions.EctoValidations","doc":"By default, the following aliases are also available for convenience:\n\n```elixir\nconfig Flint, aliases: [\n  lt: :less_than,\n  gt: :greater_than,\n  le: :less_than_or_equal_to,\n  ge: :greater_than_or_equal_to,\n  eq: :equal_to,\n  ne: :not_equal_to\n]\n```","ref":"Flint.Extensions.EctoValidations.html#module-aliases"},{"type":"module","title":"Example - Flint.Extensions.EctoValidations","doc":"```elixir\ndefmodule Person do\n  use Flint.Schema\n\n  embedded_schema do\n    field! :first_name, :string,  max: 10, min: 5\n    field! :last_name, :string, min: 5, max: 10\n    field :favorite_colors, {:array, :string}, subset_of: [\"red\", \"blue\", \"green\"]\n    field! :age, :integer, greater_than: 0, less_than: max_age\n  end\nend\n```\n\n```elixir\nPerson.changeset(\n  %Person{},\n  %{first_name: \"Bob\", last_name: \"Smith\", favorite_colors: [\"red\", \"blue\", \"pink\"], age: 101},\n  [max_age: 100]\n)\n```\n\n```elixir\n#Ecto.Changeset ,\n  valid?: false,\n  ...\n>\n```","ref":"Flint.Extensions.EctoValidations.html#module-example"},{"type":"function","title":"Flint.Extensions.EctoValidations.apply_validations/2","doc":"Applies validations to each field according to the options passed in the schema specification.\n\nSee the `Field Validations` section of the README for more information on validation details.","ref":"Flint.Extensions.EctoValidations.html#apply_validations/2"},{"type":"module","title":"Flint.Extensions.Embedded","doc":"An extension to house common default configurations for embedded schemas. These configurations are specific for\nin-memory schemas.","ref":"Flint.Extensions.Embedded.html"},{"type":"module","title":"Attributes - Flint.Extensions.Embedded","doc":"The following attributes and defaults are set by this extension:\n\n* `:schema_prefix`\n* `:schema_context`\n* `:primary_key` - defaults to `false`\n* `:timestamp_opts` - defaults to `[type: :naive_datetime]`\n\nA new schema reflection function is made for each attribute:\n\n```elixir\n__schema__(:schema_context)\n...\n```","ref":"Flint.Extensions.Embedded.html#module-attributes"},{"type":"module","title":"Flint.Extensions.JSON","doc":"Provides JSON encoding capabilities for Flint schemas with Go-like marshalling options.\n\nThis extension enhances Flint schemas with customizable JSON serialization options,\nallowing fine-grained control over how fields are represented in JSON output.","ref":"Flint.Extensions.JSON.html"},{"type":"module","title":"Usage - Flint.Extensions.JSON","doc":"To use this extension, include it in your Flint schema:\n\n```elixir\ndefmodule MySchema do\n  use Flint.Schema,\n     extensions: [{JSON, json_module: :json}]  # Jason, or Poison\n     #extensions: [JSON] # (defaults to Jason if no args passed)\n\n  embedded_schema do\n    # Schema fields...\n  end\nend\n```","ref":"Flint.Extensions.JSON.html#module-usage"},{"type":"module","title":"JSON Encoding Options - Flint.Extensions.JSON","doc":"The following options can be specified for each field in your schema:\n\n- `:name` - Specifies a custom name for the field in the JSON output.\n- `:omitempty` - When set to `true`, omits the field from JSON output if its value is `nil`.\n- `:ignore` - When set to `true`, always excludes the field from JSON output.","ref":"Flint.Extensions.JSON.html#module-json-encoding-options"},{"type":"module","title":"Defining Options - Flint.Extensions.JSON","doc":"Options are defined directly in your schema using the `field` macro:\n\n```elixir\nembedded_schema do\n  field :id, :string, name: \"ID\"\n  field :title, :string, name: \"Title\", omitempty: true\n  field :internal_data, :map, ignore: true\nend\n```","ref":"Flint.Extensions.JSON.html#module-defining-options"},{"type":"module","title":"Example - Flint.Extensions.JSON","doc":"```elixir\ndefmodule Book do\n  use Flint.Schema,\n    extensions: [Embedded, Accessible, JSON]\n\n  embedded_schema do\n    field :id, :string, name: \"ISBN\"\n    field :title, :string\n    field :author, :string, omitempty: true\n    field :price, :decimal, name: \"SalePrice\"\n    field :internal_notes, :string, ignore: true\n  end\nend\n\nbook = %Book{\n  id: \"978-3-16-148410-0\",\n  title: \"Example Book\",\n  author: nil,\n  price: Decimal.new(\"29.99\"),\n  internal_notes: \"Not for customer eyes\"\n}\n\nJason.encode!(book)\n# Results in:\n# {\n#   \"ISBN\": \"978-3-16-148410-0\",\n#   \"title\": \"Example Book\",\n#   \"SalePrice\": \"29.99\"\n# }\n```\n\nNote how the `author` field is omitted due to `omitempty: true`, the `internal_notes`\nfield is ignored, and custom names are used for `id` and `price`.","ref":"Flint.Extensions.JSON.html#module-example"},{"type":"module","title":"Introspection - Flint.Extensions.JSON","doc":"You can inspect the JSON options for a schema using the `__schema__/1` function:\n\n```elixir\nBook.__schema__(:extra_options)\n# Returns:\n# [\n#   id: [name: \"ISBN\"],\n#   title: [],\n#   author: [omitempty: true],\n#   price: [name: \"SalePrice\"],\n#   internal_notes: [ignore: true]\n# ]\n```\n\nThis allows for runtime inspection and manipulation of JSON encoding behavior.","ref":"Flint.Extensions.JSON.html#module-introspection"},{"type":"module","title":"Implementation Details - Flint.Extensions.JSON","doc":"This extension implements the `Jason.Encoder` protocol for your schema, automatically\napplying the specified options during JSON encoding. It leverages Flint's schema\nintrospection capabilities to retrieve and apply the options.\n\nThe encoding process:\n1. Converts the struct to a map\n2. Applies the `:name` option to change key names\n3. Applies the `:omitempty` option to remove nil values\n4. Applies the `:ignore` option to exclude specified fields\n5. Encodes the resulting map to JSON\n\nThis approach provides a flexible and powerful way to control JSON serialization\ndirectly from your schema definition, promoting clean and maintainable code.","ref":"Flint.Extensions.JSON.html#module-implementation-details"},{"type":"module","title":"Flint.Extensions.PostTransforms","doc":"The `PostTransforms` extension adds the `:map` option to `Flint` schemas.\n\nThis works similarly to the `PreTransforms` extension, but uses the `:map` option rather than\nthe `:derive` option used by `PreTransforms`, and by default, applies to the field after all validations.\n\nThe same caveats apply to the `:map` expression as all other expressions, with the exception that the\n`:map` function **only** accepts arity-1 anonymous functions or non-anonymous function expressions\n(eg. using variable replacement).\n\nIn the following example, `:derived` is used to normalize incoming strings to downcase to prepare for\nthe validation, then the output is mapped to the uppercase string using the `:map` option.\n\n```elixir\ndefmodule Character do\n  use Flint.Schema\n\n  embedded_schema do\n    field! :type, :string, derive: &String.downcase/1, map: String.upcase(type) do\n      type not in ~w[elf human] -> \"Expected elf or human, got: #{type}\"\n    end\n\n    field! :age, :integer do\n      age < 0 ->\n        \"Nobody can have a negative age\"\n\n      type == \"elf\" and age > max_elf_age ->\n        \"Attention! The elf has become a bug! Should be dead already!\"\n\n      type == \"human\" and age > max_human_age ->\n        \"Expected human to have up to #{max_human_age}, got: #{age}\"\n    end\n  end\nend\n```\n\n```elixir\nmax_elf_age = 400\nmax_human_age = 120\nCharacter.new!(%{type: \"Elf\", age: 10}, binding())\n```\n\n```elixir\n%Character{type: \"ELF\", age: 10}\n```","ref":"Flint.Extensions.PostTransforms.html"},{"type":"function","title":"Flint.Extensions.PostTransforms.apply_post_transforms/2","doc":"Applies transformations to each field according to the `:map` options passed in the schema specification.\n\nThese transformations are applied after validations when used within the default `Flint.Changeset.changeset` implementation.\n\nAccepts optional bindings which are passed to evaluated code.","ref":"Flint.Extensions.PostTransforms.html#apply_post_transforms/2"},{"type":"module","title":"Flint.Extensions.PreTransforms","doc":"The `PreTransforms` provides a convenient `:derive` option to express how the field is computed.\n\n**By default, this occurs after casting and before validations.**\n\n`derived` fields let you define expressions with support for custom bindings to include any\n`field` declarations that occur before the current field.\n\n`:derive` will automatically put the result of the input expression into the field value.\nBy default, this occurs before any other validation, so you can still have access to `field`\nbindings and even the current computed field value (eg. within a `:when` validation from the\n`When` extension).\n\nYou can define a `derived` field with respect to the field itself, in which case it acts as\ntransformation. Typically in `Ecto`, incoming transformations of this support would happen\nat the `cast` step, which means the behavior is determined by the type in which you are casting into.\n`:derive` lets you apply a transformation after casting to change that behavior\nwithout changing the underlying allowed type.\n\nYou can also define a `derived` field with an expression that does not depend on the field,\nin which case it is suggested that you use the `field` macro instead of `field!` since any input\nin that case would be thrashed by the derived value. This means that a field can be completely\ndetermined as a product of other fields!\n\n```elixir\ndefmodule Test do\n  use Flint.Schema\n\n  embedded_schema do\n    field! :category, Union, oneof: [Ecto.Enum, :decimal, :integer], values: [a: 1, b: 2, c: 3]\n    field! :rating, :integer, when: category == target_category\n    field :score, derive: rating + category, :integer, gt: 1, lt: 100, when: score > rating\n  end\nend\n```\n\n```elixir\nTest.new!(%{category: 1, rating: 80}, target_category: 1)\n\n# %Test{category: 1, rating: 80, score: 81}\n```","ref":"Flint.Extensions.PreTransforms.html"},{"type":"function","title":"Flint.Extensions.PreTransforms.apply_pre_transforms/2","doc":"Applies transformations to each field according to the `:derive` options passed in the schema specification.\n\nThese transformations are applied after casting, but before validations when used within the default `Flint.Changeset.changeset` implementation.\n\nAccepts optional bindings which are passed to evaluated code.","ref":"Flint.Extensions.PreTransforms.html#apply_pre_transforms/2"},{"type":"module","title":"Flint.Extensions.When","doc":"The `When` extension adds the `:when` option to `Flint` schemas.\n\n`:when` lets you define an arbitrary boolean expression that will be evaluated and pass the validation if it\nevaluates to a truthy value. You may pass bindings to this condition and\nrefer to previously defined fields. `:when` also lets you refer to the current `field` in which\nthe `:when` condition is defined. Theoretically, you could write many of the other validations using `:when`, but you will\nreceive worse error messages with `:when` than with the dedicated validations.\n\n```elixir\ndefmodule Test do\n  use Flint.Schema\n\n  embedded_schema do\n    field! :category, Union, oneof: [Ecto.Enum, :decimal, :integer], values: [a: 1, b: 2, c: 3]\n    field! :rating, :integer, when: category == target_category\n    field! :score, :integer, gt: 1, lt: 100, when: score > rating\n  end\nend\n```\n\n```elixir\n> Test.new!(%{category: :a, rating: 80, score: 10}, target_category: :a)\n\n** (ArgumentError) %Test{category: :a, rating: 80, score: [\"Failed `:when` validation\"]}\n  (flint 0.0.1) lib/schema.ex:406: Flint.Schema.new!/3\n  (elixir 1.15.7) src/elixir.erl:396: :elixir.eval_external_handler/3\n  (stdlib 5.1.1) erl_eval.erl:750: :erl_eval.do_apply/7\n  (elixir 1.15.7) src/elixir.erl:375: :elixir.eval_forms/4\n  (elixir 1.15.7) lib/module/parallel_checker.ex:112: Module.ParallelChecker.verify/1\n  lib/livebook/runtime/evaluator.ex:622: anonymous fn/3 in Livebook.Runtime.Evaluator.eval/4\n  (elixir 1.15.7) lib/code.ex:574: Code.with_diagnostics/2\n```","ref":"Flint.Extensions.When.html"},{"type":"function","title":"Flint.Extensions.When.validate_when_condition/2","doc":"","ref":"Flint.Extensions.When.html#validate_when_condition/2"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,49,46,50]}}